if (AjxPackage.define("Zimbra")) {
/*
 * ***** BEGIN LICENSE BLOCK *****
 * 
 * Zimbra Collaboration Suite Web Client
 * Copyright (C) 2006, 2007 Zimbra, Inc.
 * 
 * The contents of this file are subject to the Yahoo! Public License
 * Version 1.0 ("License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.zimbra.com/license.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * 
 * ***** END LICENSE BLOCK *****
 */
/*
 * Support for a generic login page, error dialog,
 * splash screen, and for sending requests to the server.
 */
if (AjxPackage.define("zimbra.csfe.ZmBatchCommand")) {
/*
 * ***** BEGIN LICENSE BLOCK *****
 * 
 * Zimbra Collaboration Suite Web Client
 * Copyright (C) 2006, 2007 Zimbra, Inc.
 * 
 * The contents of this file are subject to the Yahoo! Public License
 * Version 1.0 ("License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.zimbra.com/license.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * 
 * ***** END LICENSE BLOCK *****
 */

/**
 * Creates an empty batch command. Use its run() method to add commands to it.
 * @constructor
 * @class
 * This class represent a batch command, which is a collection of separate SOAP
 * requests. Each command is a callback with a method, arguments, and (usually) an
 * object on which to call the method. Normally, when the command is run, it creates
 * a soap document which it hands to the app controller's sendRequest() method. It may
 * also pass a response callback and/or an error callback.
 * <p>
 * Instead of calling sendRequest(), the command should hand the batch command its SOAP
 * document, response callback, and error callback. The last argument that the command
 * receives is a reference to the batch command; that's how it knows it's in batch mode.
 * </p><p>
 * After all commands have been added to the batch command, call its run() method. That will
 * create a BatchRequest out of the individual commands' SOAP documents and send it to the
 * server. Each subrequest gets an ID. When the BatchResponse comes back, it is broken into
 * individual responses. If a response indicates success (it is a *Response), the corresponding
 * response callback is called with the result. If the response is a fault, the corresponding
 * error callback is called with the exception.
 * </p><p>
 * A command does not have to be the method that generates a SOAP document. It can be higher-level.
 * Just make sure that the reference to the batch command gets passed down to it.
 * </p>
 * @author Conrad Damon
 * 
 * @param continueOnError	[boolean]*	if true, the batch request continues processing
 * 										when a subrequest fails (defaults to true)
 * @param accountName		[string]	The account name to run this batch command as.
 */
ZmBatchCommand = function(continueOnError, accountName) {
	
	this._onError = (continueOnError === false) ? ZmBatchCommand.STOP : ZmBatchCommand.CONTINUE;
	this._accountName = accountName;

	this.curId = 0;
    this._cmds = [];
	this._soapDocs = [];
	this._respCallbacks = [];
	this._errorCallbacks = [];
	this._execFrames = [];
};

ZmBatchCommand.prototype.toString =
function() {
	return "ZmBatchCommand";
};

//
// Constants
//

ZmBatchCommand.STOP = "stop";
ZmBatchCommand.CONTINUE = "continue";

/**
 * Adds a command to the list of commands to run as part of this batch request.
 * 
 * @param cmd	[AjxCallback]	a command
 */
ZmBatchCommand.prototype.add =
function(cmd) {
	this._cmds.push(cmd);
};

ZmBatchCommand.prototype.size =
function() {
	return this.curId;
};

/**
 * Issues the batch request. For each individual request, either a response or an
 * error callback will be called.
 * 
 * @param callback		[AjxCallback]*	callback to run after entire batch request has completed
 * @param errorCallback	[AjxCallback]*	Error callback called if anything fails.
 *										The error callbacks arguments are all
 *										of the exceptions that occured. Note:
 *										only the first exception is passed if
 *										this batch command's onError is set to
 *										stop.
 */
ZmBatchCommand.prototype.run =
function(callback, errorCallback) {

	// Invoke each command so that it hands us its SOAP doc, response callback, and
	// error callback
    for (var i = 0; i < this._cmds.length; i++) {
		var cmd = this._cmds[i];
		cmd.run(this);
        this.curId++;
	}

	// Create the BatchRequest
	var batchSoapDoc = AjxSoapDoc.create("BatchRequest", "urn:zimbra");
	batchSoapDoc.setMethodAttribute("onerror", this._onError);

	// Add each command's request element to the BatchRequest, and set its ID
    var size = this.size();
    for (var i = 0; i < size; i++) {
		var soapDoc = this._soapDocs[i];
		var reqEl = soapDoc.getMethod();
		reqEl.setAttribute("requestId", i);
		var node = batchSoapDoc.adoptNode(reqEl);
		batchSoapDoc.getMethod().appendChild(node);
	}
	
	// Issue the BatchRequest
	var params = {
		soapDoc:		batchSoapDoc,
		asyncMode:		true,
		callback:		new AjxCallback(this, this._handleResponseRun, [callback, errorCallback]),
		errorCallback:	errorCallback,
		accountName:	this._accountName
	};
	appCtxt.getAppController().sendRequest(params);
};

ZmBatchCommand.prototype._handleResponseRun =
function(callback, errorCallback, result) {
	var batchResponse = result.getResponse();
	if (!batchResponse.BatchResponse) {
		DBG.println(AjxDebug.DBG1, "Missing batch response!");
		return;
	}
	// NOTE: In case the order of the requests is significant, we process
	//       the responses in the same order.
	var responses = [];
	for (var method in batchResponse.BatchResponse) {
		if (method.match(/^_/)) continue;

		var methodResponses = batchResponse.BatchResponse[method];
		for (var i = 0; i < methodResponses.length; i++) {
			responses[methodResponses[i].requestId] = { method: method, resp: methodResponses[i] };
		}
	}
	var exceptions = [];
	for (var i = 0; i < responses.length; i++) {
		var response = responses[i];
		try {
			this._processResponse(response.method, response.resp);
		}
		catch (ex) {
			exceptions.push(ex);
			if (this._onError == ZmBatchCommand.STOP) {
				break;
			}
		}
	}
	if (exceptions.length > 0 && errorCallback) {
		errorCallback.run.apply(errorCallback, exceptions);
	}
	else if (callback) {
		callback.run(result);
	}
};

ZmBatchCommand.prototype._handleResponseRunSafari =
function(callback, result) {
	var resp = result.getResponse();
	for (var i in resp) {
		this._processResponse(i, resp[i]);
	}

	// only run the final callback once all async requests have returned
    var size = this.size();
    if (++this._responseCount == size && callback) {
		callback.run(result);
	}
};

/**
 * Adds the given command parameters to the batch command, as part of a command's
 * invocation.
 * 
 * @param soapDoc		[AjxSoapDoc]	a SOAP document with the command's request
 * @param callback		[AjxCallback]*	next callback in chain for async request
 * @param errorCallback	[Object]*		callback to run if there is an exception
 * @param execFrame		[AjxCallback]*	the calling method, object, and args
 */
ZmBatchCommand.prototype.addRequestParams =
function(soapDoc, respCallback, errorCallback, execFrame) {
	this._soapDocs[this.curId] = soapDoc;
	this._respCallbacks[this.curId] = respCallback;
	this._errorCallbacks[this.curId] = errorCallback;
	this._execFrames[this.curId] = execFrame;
};

ZmBatchCommand.prototype.addNewRequestParams =
function(soapDoc, respCallback, errorCallback, execFrame) {
    this.addRequestParams(soapDoc, respCallback, errorCallback, execFrame);
    this.curId++;
};

/*
 * Each type of request will return an array of *Response elements. There may also be
 * an array of Fault elements. Each element has an ID, so we can match it to its
 * response or error callback, and run whichever is appropriate.
 */
ZmBatchCommand.prototype._processResponse =
function(method, resp) {
	var id = resp.requestId;

	// handle error
	if (method == "Fault") {
		var ex = ZmCsfeCommand.faultToEx(resp, "ZmBatchCommand.prototype.run");
		var execFrame = this._execFrames[id];
		if (this._errorCallbacks[id]) {
			var handled = this._errorCallbacks[id].run(ex);
			if (!handled && execFrame) {
				appCtxt.getAppController()._handleException(ex, execFrame);
			}
		} else if (execFrame) {
			appCtxt.getAppController()._handleException(ex, execFrame);
		}
		throw ex;
	}

	// process response callback
	if (this._respCallbacks[id]) {
		var data = {};
		data[method] = resp;
		var result = new ZmCsfeResult(data);
		this._respCallbacks[id].run(result, resp);
	}
};
}
if (AjxPackage.define("zimbra.csfe.ZmCsfeCommand")) {
/*
 * ***** BEGIN LICENSE BLOCK *****
 * 
 * Zimbra Collaboration Suite Web Client
 * Copyright (C) 2004, 2005, 2006, 2007 Zimbra, Inc.
 * 
 * The contents of this file are subject to the Yahoo! Public License
 * Version 1.0 ("License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.zimbra.com/license.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * 
 * ***** END LICENSE BLOCK *****
 */

ZmCsfeCommand = function() {
};

// Static properties

// Global settings for each CSFE command
ZmCsfeCommand._COOKIE_NAME = "ZM_AUTH_TOKEN";
ZmCsfeCommand.serverUri = null;
ZmCsfeCommand._sessionId = null;

// Static methods

ZmCsfeCommand.getAuthToken =
function() {
	return AjxCookie.getCookie(document, ZmCsfeCommand._COOKIE_NAME)
};

ZmCsfeCommand.setCookieName =
function(cookieName) {
	ZmCsfeCommand._COOKIE_NAME = cookieName;
};

ZmCsfeCommand.setServerUri =
function(uri) {
	ZmCsfeCommand.serverUri = uri;
};

ZmCsfeCommand.setAuthToken =
function(authToken, lifetimeMs, sessionId) {
	if (lifetimeMs != null) {
		var exp = null;
		if(lifetimeMs > 0) {
			exp = new Date();
			var lifetime = parseInt(lifetimeMs);
			exp.setTime(exp.getTime() + lifetime);
		}
		AjxCookie.setCookie(document, ZmCsfeCommand._COOKIE_NAME, authToken, exp, "/");		
	} else {
		AjxCookie.deleteCookie(document, ZmCsfeCommand._COOKIE_NAME, "/");
	}
	if (sessionId) {
		ZmCsfeCommand.setSessionId(sessionId);
	}
};

ZmCsfeCommand.clearAuthToken =
function() {
	AjxCookie.deleteCookie(document, ZmCsfeCommand._COOKIE_NAME, "/");
};

ZmCsfeCommand.getSessionId =
function() {
	return ZmCsfeCommand._sessionId;
};

ZmCsfeCommand.setSessionId =
function(sessionId) {
	var id = (sessionId instanceof Array) ? sessionId[0].id : sessionId;
	ZmCsfeCommand._sessionId = parseInt(id);
};

ZmCsfeCommand.faultToEx =
function(fault, method) {
	var faultCode = AjxStringUtil.getAsString(fault.Code.Value);
	var errorCode = AjxStringUtil.getAsString(fault.Detail.Error.Code);
	var msg = AjxStringUtil.getAsString(fault.Reason.Text);
	return new ZmCsfeException(msg, errorCode, method, faultCode, fault.Detail.Error.a);
};

/**
 * Sends a SOAP request to the server and processes the response.
 *
 * @param params			[hash]			hash of params:
 *        soapDoc			[AjxSoapDoc]	The SOAP document that represents the request
 *        noAuthToken		[boolean]*		If true, the check for an auth token is skipped
 *        serverUri			[string]*		URI to send the request to
 *        targetServer		[string]*		Host that services the request
 *        useXml			[boolean]*		If true, an XML response is requested
 *        noSession			[boolean]*		If true, no session info is included
 *        changeToken		[string]*		Current change token
 *        highestNotifySeen [int]*  	     Sequence # of the highest notification we have processed
 *        asyncMode			[boolean]*		If true, request sent asynchronously
 *        callback			[AjxCallback]*	Callback to run when response is received (async mode)
 *        logRequest		[boolean]*		If true, SOAP command name is appended to server URL
 *        accountId			[string]*		ID of account to execute on behalf of
 *        accountName		[string]*		name of account to execute on behalf of
 *        skipAuthCheck		[boolean]*		don't check if auth token has changed
 */
ZmCsfeCommand.prototype.invoke =
function(params) {

	if (!params.soapDoc) return;

	var soapDoc = params.soapDoc;
	// Add the SOAP header and context
	var hdr = soapDoc.createHeaderElement();
	var context = soapDoc.set("context", null, hdr, "urn:zimbra");

	var ua = soapDoc.set("userAgent", null, context);
	var name = ["ZimbraWebClient - ", AjxEnv.browser, " (", AjxEnv.platform, ")"].join("");
	ua.setAttribute("name", name);
	if (ZmCsfeCommand.clientVersion) {
		ua.setAttribute("version", ZmCsfeCommand.clientVersion);
	}

	if (params.noSession) {
		soapDoc.set("nosession", null, context);
	}
	var sessionId = ZmCsfeCommand.getSessionId();
	if (sessionId) {
		var si = soapDoc.set("sessionId", null, context);
		si.setAttribute("id", sessionId);
	}
	if (params.targetServer) {
		soapDoc.set("targetServer", params.targetServer, context);
	}
	if (params.highestNotifySeen) {
	  	var notify = soapDoc.set("notify", null, context);
	  	notify.setAttribute("seq", params.highestNotifySeen);
	}
	if (params.changeToken) {
		var ct = soapDoc.set("change", null, context);
		ct.setAttribute("token", params.changeToken);
		ct.setAttribute("type", "new");
	}

	// if we're not checking auth token, we don't want token/acct mismatch	
	if (!params.skipAuthCheck) {
		if (params.accountId) {
			var acc = soapDoc.set("account", params.accountId, context);
			acc.setAttribute("by", "id");
		} else if (params.accountName) {
			var acc = soapDoc.set("account", params.accountName, context);
			acc.setAttribute("by", "name");
		}
	}
	
	// Get auth token from cookie if required
	if (!params.noAuthToken) {
		var authToken = ZmCsfeCommand.getAuthToken();
		if (!authToken) {
			throw new ZmCsfeException("AuthToken required", ZmCsfeException.NO_AUTH_TOKEN, "ZmCsfeCommand.invoke");
		}
		if (ZmCsfeCommand._curAuthToken && !params.skipAuthCheck && (authToken != ZmCsfeCommand._curAuthToken)) {
			throw new ZmCsfeException("AuthToken has changed", ZmCsfeException.AUTH_TOKEN_CHANGED, "ZmCsfeCommand.invoke");
		}
		soapDoc.set("authToken", authToken, context);
		ZmCsfeCommand._curAuthToken = authToken;
	}
	
	// Tell server what kind of response we want
	if (!params.useXml) {
		var js = soapDoc.set("format", null, context);
		js.setAttribute("type", "js");
	}

	var asyncMode = params.asyncMode;
	var methodNameStr = soapDoc.getMethod().nodeName;
	DBG.println(AjxDebug.DBG1, ["<H4>", methodNameStr, (asyncMode) ? " (asynchronous)" : "" ,"</H4>"].join(""), methodNameStr);
	DBG.printXML(AjxDebug.DBG1, soapDoc.getXml());

	var rpcCallback;
	try {
		var uri = params.serverUri || ZmCsfeCommand.serverUri;
		//if (params.logRequest)
			uri = uri + soapDoc._methodEl.nodeName;
		var requestStr = soapDoc.getXml();

		this._st = new Date();

		if (asyncMode) {
//			DBG.println(AjxDebug.DBG1, "set callback for asynchronous response");	
			rpcCallback = new AjxCallback(this, this._runCallback, params.callback);
			this._rpcId = AjxRpc.invoke(requestStr, uri, {"Content-Type": "application/soap+xml; charset=utf-8"}, rpcCallback);
		} else {
//			DBG.println(AjxDebug.DBG1, "parse response synchronously");	
			var response = AjxRpc.invoke(requestStr, uri, {"Content-Type": "application/soap+xml; charset=utf-8"});
			if (!params.returnXml) {
				return this._getResponseData(response, false);
			} else {
				return response;
			}
		}
	} catch (ex) {
		if (!(ex && (ex instanceof ZmCsfeException || ex instanceof AjxSoapException || ex instanceof AjxException))) {
			var newEx = new ZmCsfeException();
			newEx.method = "ZmCsfeCommand.invoke";
			newEx.detail = ex ? ex.toString() : "undefined exception";
			newEx.code = ZmCsfeException.UNKNOWN_ERROR;
			newEx.msg = "Unknown Error";
			ex = newEx;
		}
		if (asyncMode) {
			rpcCallback.run(new ZmCsfeResult(ex, true));
		} else {
			throw ex;
		}
	}
};

/**
* Takes the response to an RPC request and returns a JS object with the response data.
*
* @param response	[Object]	RPC response with properties "text" and "xml"
* @param asyncMode	[boolean]	true if we're in asynchronous mode
*/
ZmCsfeCommand.prototype._getResponseData =
function(response, asyncMode) {
	this._en = new Date();
	DBG.println(AjxDebug.DBG1, "ROUND TRIP TIME: " + (this._en.getTime() - this._st.getTime()));

	var result = new ZmCsfeResult();
	var xmlResponse = false;
	var respDoc = null;
	if (typeof(response.text) == "string" && response.text.indexOf("{") == 0) {
		respDoc = response.text;
	} else {
		// an XML response if we requested one, or a fault
		try {
			xmlResponse = true;
			if (!(response.text || (response.xml && (typeof response.xml) == "string"))) {
				// If we can't reach the server, req returns immediately with an empty response rather than waiting and timing out
				throw new ZmCsfeException(null, ZmCsfeException.EMPTY_RESPONSE, "ZmCsfeCommand.prototype.invoke");
			}
			// responseXML is empty under IE
			respDoc = (AjxEnv.isIE || response.xml == null) ? AjxSoapDoc.createFromXml(response.text) :
															  AjxSoapDoc.createFromDom(response.xml);
		} catch (ex) {
			DBG.dumpObj(AjxDebug.DBG1, ex);
			if (asyncMode) {
				result.set(ex, true);
				return result;
			} else {
				throw ex;
			}
		}
		if (!respDoc) {
			var ex = new ZmCsfeException(null, ZmCsfeException.SOAP_ERROR, "ZmCsfeCommand.prototype.invoke", "Bad XML response doc");
			DBG.dumpObj(AjxDebug.DBG1, ex);
			if (asyncMode) {
				result.set(ex, true);
				return result;
			} else {
				throw ex;
			}
		}
	}

	var linkName = "Response";
	if (respDoc && respDoc.match) {
		var m = respDoc.match(/\{"?Body"?:\{"?(\w+)"?:/);
		if (m && m.length) linkName = m[1];
	}
	DBG.println(AjxDebug.DBG1, ["<H4> RESPONSE", (asyncMode) ? " (asynchronous)" : "" ,"</H4>"].join(""), linkName);

	var obj = {};

	if (xmlResponse) {
		DBG.printXML(AjxDebug.DBG1, respDoc.getXml());
		obj = respDoc._xmlDoc.toJSObject(true, false, true);
	} else {
		try {
			eval("obj=" + respDoc);
		} catch (ex) {
			if (ex.name == "SyntaxError") {
				ex = new ZmCsfeException(null, ZmCsfeException.BAD_JSON_RESPONSE, "ZmCsfeCommand.prototype._getResponseData");
			}
			DBG.dumpObj(AjxDebug.DBG1, ex);
			if (asyncMode) {
				result.set(ex, true);
				return result;
			} else {
				throw ex;
			}
		}

	}

	DBG.dumpObj(AjxDebug.DBG1, obj, -1);

	var fault = obj.Body.Fault;
	if (fault) {
		// JS response with fault
		var ex = ZmCsfeCommand.faultToEx(fault, "ZmCsfeCommand.prototype.invoke");
		if (asyncMode) {
			result.set(ex, true, obj.Header);
			return result;
		} else {
			throw ex;
		}
	} else if (!response.success) {
		// bad XML or JS response that had no fault
		var ex = new ZmCsfeException(null, ZmCsfeException.CSFE_SVC_ERROR,
									 "ZmCsfeCommand.prototype.invoke", "HTTP response status " + response.status);
		if (asyncMode) {
			result.set(ex, true);
			return result;
		} else {
			throw ex;
		}
	} else {
		// good response
		if (asyncMode) {
			result.set(obj);
		}
	}

	if (obj.Header && obj.Header.context && obj.Header.context.sessionId) {
		ZmCsfeCommand.setSessionId(obj.Header.context.sessionId);
	}

	return asyncMode ? result : obj;
};

/**
* Runs the callback that was passed to invoke() for an async command.
*
* @param callback	[AjxCallback]	Callback to run with response data
* @param response	[Object]		RPC response object
*/
ZmCsfeCommand.prototype._runCallback =
function(callback, result) {
	if (!result) return;

	var response;
	if (result instanceof ZmCsfeResult) {
		response = result; // we already got an exception and packaged it
	} else {
		response = this._getResponseData(result, true);
	}
	this._en = new Date();

	if (!callback) {
		DBG.println(AjxDebug.DBG1, "Could not find callback!");
		return;
	}

	if (callback) callback.run(response);
};

/**
* Cancels this request (which must be async).
*/
ZmCsfeCommand.prototype.cancel =
function() {
	if (!this._rpcId) return;

	var req = AjxRpc.getRpcRequest(this._rpcId);
	if (req)
		req.cancel();
};

// DEPRECATED - instead, use instance method invoke() above
ZmCsfeCommand.invoke =
function(soapDoc, noAuthTokenRequired, serverUri, targetServer, useXml, noSession, changeToken) {
	var hdr = soapDoc.createHeaderElement();
	var context = soapDoc.set("context", null, hdr, "urn:zimbra");

	if (noSession)
		soapDoc.set("nosession", null, context);
	var sessionId = ZmCsfeCommand.getSessionId();
	if (sessionId) {
		var si = soapDoc.set("sessionId", null, context);
		si.setAttribute("id", sessionId);
	}
	if (targetServer)
		soapDoc.set("targetServer", targetServer, context);
	if (changeToken) {
		var ct = soapDoc.set("change", null, context);
		ct.setAttribute("token", changeToken);
		ct.setAttribute("type", "new");
	}

	// See if we have an auth token, if not, then mock up and need to authenticate or just have no auth cookie
	if (!noAuthTokenRequired) {
		var authToken = ZmCsfeCommand.getAuthToken();
		if (!authToken)
			throw new ZmCsfeException("AuthToken required", ZmCsfeException.NO_AUTH_TOKEN, "ZmCsfeCommand.invoke");
		soapDoc.set("authToken", authToken, context);
	}

	if (!useXml) {
		var js = soapDoc.set("format", null, context);
		js.setAttribute("type", "js");
	}

	DBG.println(AjxDebug.DBG1, "<H4>REQUEST</H4>");
	DBG.printXML(AjxDebug.DBG1, soapDoc.getXml());

	var xmlResponse = false;
	try {
		var uri = serverUri || ZmCsfeCommand.serverUri;
		var requestStr = soapDoc.getXml();
			
		var _st = new Date();
		var response = AjxRpc.invoke(requestStr, uri, {"Content-Type": "application/soap+xml; charset=utf-8"});
		var _en = new Date();
		DBG.println(AjxDebug.DBG1, "ROUND TRIP TIME: " + (_en.getTime() - _st.getTime()));

		var respDoc = null;
		if (typeof(response.text) == "string" && response.text.indexOf("{") == 0) {
			respDoc = response.text;
		} else {
			xmlResponse = true;
			// responseXML is empty under IE
			respDoc = (AjxEnv.isIE || response.xml == null)
				? AjxSoapDoc.createFromXml(response.text) 
				: AjxSoapDoc.createFromDom(response.xml);
		}
	} catch (ex) {
		if (ex instanceof AjxSoapException) {
			throw ex;
		} else if (ex instanceof AjxException) {
			throw ex; 
		}  else {
			var newEx = new ZmCsfeException();
			newEx.method = "ZmCsfeCommand.invoke";
			newEx.detail = ex.toString();
			newEx.code = ZmCsfeException.UNKNOWN_ERROR;
			newEx.msg = "Unknown Error";
			throw newEx;
		}
	}
	
	DBG.println(AjxDebug.DBG1, "<H4>RESPONSE</H4>");

	var resp;
	if (xmlResponse) {
		DBG.printXML(AjxDebug.DBG1, respDoc.getXml());
		var body = respDoc.getBody();
		var fault = AjxSoapDoc.element2FaultObj(body);
		if (fault) {
			throw new ZmCsfeException("Csfe service error", fault.errorCode, "ZmCsfeCommand.invoke", fault.reason);
		}
		if (useXml)
			return body;

		resp = "{";
		var hdr = respDoc.getHeader();
		if (hdr)
			resp += AjxUtil.xmlToJs(hdr) + ",";
		resp += AjxUtil.xmlToJs(body);
		resp += "}";
	} else {
		resp = respDoc;	
	}

	var obj = new Object();
	eval("obj=" + resp);
	DBG.dumpObj(obj, -1);

	var fault = obj.Body.Fault;
	if (fault)
		throw new ZmCsfeException(fault.Reason.Text, fault.Detail.Error.Code, "ZmCsfeCommand.invoke", fault.Code.Value);
	if (obj.Header && obj.Header.context && obj.Header.context.sessionId)
		ZmCsfeCommand.setSessionId(obj.Header.context.sessionId);

	return obj;
};
}
if (AjxPackage.define("zimbra.csfe.ZmCsfeException")) {
/*
 * ***** BEGIN LICENSE BLOCK *****
 * 
 * Zimbra Collaboration Suite Web Client
 * Copyright (C) 2004, 2005, 2006, 2007 Zimbra, Inc.
 * 
 * The contents of this file are subject to the Yahoo! Public License
 * Version 1.0 ("License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.zimbra.com/license.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * 
 * ***** END LICENSE BLOCK *****
 */

ZmCsfeException = function(msg, code, method, detail, data) {
	AjxException.call(this, msg, code, method, detail);
	
	if (data) {
		this.data = {};
		for (var i = 0; i < data.length; i++) {
			var item = data[i];
			var key = item.n;
			if (!this.data[key]) {
				this.data[key] = [];
			}
			this.data[key].push(item._content);
		}
	}
};

ZmCsfeException.prototype = new AjxException;
ZmCsfeException.prototype.constructor = ZmCsfeException;

ZmCsfeException.prototype.toString =
function() {
	return "ZmCsfeException";
};

//
// Constants
//

// structured data keys
ZmCsfeException.MAIL_SEND_ADDRESS_FAILURE_INVALID = "invalid";
ZmCsfeException.MAIL_SEND_ADDRESS_FAILURE_UNSENT = "unsent";

//
// Static functions
//

ZmCsfeException.getErrorMsg =
function(code, args) {
	var msg = ZMsg[code];
	if (!msg) { return ""; }
	this.msg = this.msg || msg;
	return args ? AjxMessageFormat.format(msg, args) : msg;
};

//
// Public methods
//

ZmCsfeException.prototype.getErrorMsg =
function(args) {
	return ZmCsfeException.getErrorMsg(this.code, args);
};

ZmCsfeException.prototype.getData =
function(key) {
	return this.data ? this.data[key] : null;
};

//
// Constants for server exceptions
//

ZmCsfeException.AUTH_TOKEN_CHANGED	= "AUTH_TOKEN_CHANGED";
ZmCsfeException.BAD_JSON_RESPONSE	= "BAD_JSON_RESPONSE";
ZmCsfeException.CSFE_SVC_ERROR		= "CSFE_SVC_ERROR";
ZmCsfeException.EMPTY_RESPONSE		= "EMPTY_RESPONSE";
ZmCsfeException.NETWORK_ERROR		= "NETWORK_ERROR";
ZmCsfeException.NO_AUTH_TOKEN		= "NO_AUTH_TOKEN";
ZmCsfeException.SOAP_ERROR			= "SOAP_ERROR";

ZmCsfeException.LICENSE_ERROR				= "service.LICENSE_ERROR";
ZmCsfeException.SVC_ALREADY_IN_PROGRESS		= "service.ALREADY_IN_PROGRESS";
ZmCsfeException.SVC_AUTH_EXPIRED			= "service.AUTH_EXPIRED";
ZmCsfeException.SVC_AUTH_REQUIRED			= "service.AUTH_REQUIRED";
ZmCsfeException.SVC_FAILURE					= "service.FAILURE";
ZmCsfeException.SVC_INVALID_REQUEST			= "service.INVALID_REQUEST";
ZmCsfeException.SVC_PARSE_ERROR				= "service.PARSE_ERROR";
ZmCsfeException.SVC_PERM_DENIED				= "service.PERM_DENIED";
ZmCsfeException.SVC_RESOURCE_UNREACHABLE	= "service.RESOURCE_UNREACHABLE";
ZmCsfeException.SVC_UNKNOWN_DOCUMENT		= "service.UNKNOWN_DOCUMENT";
ZmCsfeException.SVC_TEMPORARILY_UNAVAILABLE	= "service.TEMPORARILY_UNAVAILABLE";
ZmCsfeException.SVC_WRONG_HOST				= "service.WRONG_HOST";

ZmCsfeException.ACCT_AUTH_FAILED			= "account.AUTH_FAILED";
ZmCsfeException.ACCT_CHANGE_PASSWORD		= "account.CHANGE_PASSWORD";
ZmCsfeException.ACCT_EXISTS					= "account.ACCOUNT_EXISTS";
ZmCsfeException.ACCT_INVALID_PASSWORD		= "account.INVALID_PASSWORD";
ZmCsfeException.ACCT_INVALID_PREF_NAME		= "account.INVALID_PREF_NAME";
ZmCsfeException.ACCT_INVALID_PREF_VALUE		= "account.INVALID_PREF_VALUE";
ZmCsfeException.ACCT_MAINTENANCE_MODE		= "account.MAINTENANCE_MODE";
ZmCsfeException.ACCT_NO_SUCH_ACCOUNT		= "account.NO_SUCH_ACCOUNT";
ZmCsfeException.ACCT_NO_SUCH_SAVED_SEARCH	= "account.NO_SUCH_SAVED_SEARCH";
ZmCsfeException.ACCT_NO_SUCH_TAG			= "account.ACCT_NO_SUCH_TAG";
ZmCsfeException.ACCT_PASS_CHANGE_TOO_SOON	= "account.PASSWORD_CHANGE_TOO_SOON";
ZmCsfeException.ACCT_PASS_LOCKED			= "account.PASSWORD_LOCKED";
ZmCsfeException.ACCT_PASS_RECENTLY_USED		= "account.PASSWORD_RECENTLY_USED";
ZmCsfeException.COS_EXISTS					= "account.COS_EXISTS";
ZmCsfeException.DISTRIBUTION_LIST_EXISTS	= "account.DISTRIBUTION_LIST_EXISTS";
ZmCsfeException.DOMAIN_EXISTS				= "account.DOMAIN_EXISTS";
ZmCsfeException.DOMAIN_NOT_EMPTY			= "account.DOMAIN_NOT_EMPTY";
ZmCsfeException.IDENTITY_EXISTS				= "account.IDENTITY_EXISTS";
ZmCsfeException.NO_SUCH_DISTRIBUTION_LIST	= "account.NO_SUCH_DISTRIBUTION_LIST";
ZmCsfeException.NO_SUCH_DOMAIN				= "account.NO_SUCH_DOMAIN";
ZmCsfeException.MAINTENANCE_MODE			= "account.MAINTENANCE_MODE";
ZmCsfeException.TOO_MANY_IDENTITIES			= "account.TOO_MANY_IDENTITIES";
ZmCsfeException.TOO_MANY_SEARCH_RESULTS		= "account.TOO_MANY_SEARCH_RESULTS";

ZmCsfeException.MAIL_ALREADY_EXISTS					= "mail.ALREADY_EXISTS";
ZmCsfeException.MAIL_IMMUTABLE						= "mail.IMMUTABLE_OBJECT";
ZmCsfeException.MAIL_INVALID_NAME					= "mail.INVALID_NAME";
ZmCsfeException.MAIL_MAINTENANCE_MODE				= "mail.MAINTENANCE_MODE";
ZmCsfeException.MAIL_MESSAGE_TOO_BIG				= "mail.MESSAGE_TOO_BIG";
ZmCsfeException.MAIL_NO_SUCH_CONV					= "mail.NO_SUCH_CONV";
ZmCsfeException.MAIL_NO_SUCH_FOLDER					= "mail.NO_SUCH_FOLDER";
ZmCsfeException.MAIL_NO_SUCH_ITEM					= "mail.NO_SUCH_ITEM";
ZmCsfeException.MAIL_NO_SUCH_MSG					= "mail.NO_SUCH_MSG";
ZmCsfeException.MAIL_NO_SUCH_PART					= "mail.NO_SUCH_PART";
ZmCsfeException.MAIL_NO_SUCH_TAG					= "mail.NO_SUCH_TAG";
ZmCsfeException.MAIL_QUERY_PARSE_ERROR				= "mail.QUERY_PARSE_ERROR";
ZmCsfeException.MAIL_QUOTA_EXCEEDED					= "mail.QUOTA_EXCEEDED";
ZmCsfeException.MAIL_SEND_ABORTED_ADDRESS_FAILURE	= "mail.SEND_ABORTED_ADDRESS_FAILURE";
ZmCsfeException.MAIL_SEND_FAILURE					= "mail.SEND_FAILURE";
ZmCsfeException.MAIL_TOO_MANY_CONTACTS				= "mail.TOO_MANY_CONTACTS";
ZmCsfeException.MAIL_TOO_MANY_TERMS					= "mail.TOO_MANY_QUERY_TERMS_EXPANDED";
ZmCsfeException.MAIL_UNABLE_TO_IMPORT_APPOINTMENTS	= "mail.MAIL_UNABLE_TO_IMPORT_APPOINTMENTS";
ZmCsfeException.MAIL_UNABLE_TO_IMPORT_CONTACTS		= "mail.UNABLE_TO_IMPORT_CONTACTS";
ZmCsfeException.MODIFY_CONFLICT						= "mail.MODIFY_CONFLICT";
ZmCsfeException.TOO_MANY_TAGS						= "mail.TOO_MANY_TAGS";

ZmCsfeException.VOLUME_NO_SUCH_PATH = "volume.NO_SUCH_PATH";
}
if (AjxPackage.define("zimbra.csfe.ZmCsfeResult")) {
/*
 * ***** BEGIN LICENSE BLOCK *****
 * 
 * Zimbra Collaboration Suite Web Client
 * Copyright (C) 2005, 2006, 2007 Zimbra, Inc.
 * 
 * The contents of this file are subject to the Yahoo! Public License
 * Version 1.0 ("License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.zimbra.com/license.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * 
 * ***** END LICENSE BLOCK *****
 */

/**
* Creates a CSFE result object.
* @constructor
* @class
* This class represents the result of a CSFE request. The data is either the 
* response that was received, or an exception. If the request resulted in a 
* SOAP fault from the server, there will also be a SOAP header present.
*
* @author Conrad Damon
* @param data			[Object]	response data
* @param isException	[boolean]	true if the data is an exception object
* @param header			[object]	the SOAP header
*/
ZmCsfeResult = function(data, isException, header) {
	this.set(data, isException, header);
};

ZmCsfeResult.prototype.toString =
function() {
	return "ZmCsfeResult";
};

/**
* Sets the content of the result.
*
* @param data			[Object]	response data
* @param isException	[boolean]	true if the data is an exception object
* @param header			[object]	the SOAP header
*/
ZmCsfeResult.prototype.set =
function(data, isException, header) {
	this._data = data;
	this._isException = (isException === true);
	this._header = header;
};

/**
* Returns the response data. If there was an exception, throws the exception.
*/
ZmCsfeResult.prototype.getResponse =
function() {
	if (this._isException)
		throw this._data;
	else
		return this._data;
};

/**
* Returns the exception object, if any.
*/
ZmCsfeResult.prototype.getException =
function() {
	return this._isException ? this._data : null;
};

ZmCsfeResult.prototype.isException = 
function() {
	return this._isException;
};

/**
* Returns the SOAP header that came with a SOAP fault.
*/
ZmCsfeResult.prototype.getHeader =
function() {
	return this._header;
};
}
if (AjxPackage.define("zimbra.common.ZmBaseSplashScreen")) {
/*
 * ***** BEGIN LICENSE BLOCK *****
 * 
 * Zimbra Collaboration Suite Web Client
 * Copyright (C) 2005, 2006, 2007 Zimbra, Inc.
 * 
 * The contents of this file are subject to the Yahoo! Public License
 * Version 1.0 ("License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.zimbra.com/license.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * 
 * ***** END LICENSE BLOCK *****
 */

ZmBaseSplashScreen = function(shell, imageInfo, className) {

 	if (arguments.length == 0) return;
	
 	if (!(shell instanceof DwtShell)) {
 		throw new AjxException("Parent must be a DwtShell", AjxException.INVALIDPARENT, "ZSplashScreen");
 	}
	
 	className = className || "ZSplashScreen";
 	DwtControl.call(this, shell, className, Dwt.ABSOLUTE_STYLE);

	this.__createContents();
}

ZmBaseSplashScreen.prototype = new DwtControl;
ZmBaseSplashScreen.prototype.constructor = ZmBaseSplashScreen;

/** abstract **/
ZmBaseSplashScreen.prototype.getHtml = function() { }

ZmBaseSplashScreen.prototype.setVisible =
function(visible) {
	if (visible == this.getVisible()) {
		return;
	}
	
	if (visible) {
		this.__createContents();
	}		

	DwtControl.prototype.setVisible.call(this, visible);	
	
	if (!visible) {
		this.getHtmlElement().innerHTML = "";
	}
};

ZmBaseSplashScreen.prototype.__createContents =
function() {
	var htmlEl = this.getHtmlElement();
 	htmlEl.style.zIndex = Dwt.Z_SPLASH;
	
 	var myTable = document.createElement("table");
 	myTable.border = myTable.cellSpacing = myTable.cellPadding = 0;
 	Dwt.setSize(myTable, "100%", "100%");
	
 	var row = myTable.insertRow(0);
 	var cell = row.insertCell(0);
 	cell.vAlign = "middle";
 	cell.align = "center";
	cell.innerHTML = this.getHtml();
 	htmlEl.appendChild(myTable);
	htmlEl.style.cursor = "wait";
};
}
if (AjxPackage.define("zimbra.common.ZmErrorDialog")) {
/*
 * ***** BEGIN LICENSE BLOCK *****
 * 
 * Zimbra Collaboration Suite Web Client
 * Copyright (C) 2005, 2006, 2007 Zimbra, Inc.
 * 
 * The contents of this file are subject to the Yahoo! Public License
 * Version 1.0 ("License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.zimbra.com/license.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * 
 * ***** END LICENSE BLOCK *****
 */
/**
 * Creates an error dialog which basically means it will have a "Report" button
 * @constructor
 * @class
 * A normal DwtMessageDialog w/ a "Report" button that will post user info to the 
 * server when clicked.
 */
ZmErrorDialog = function(parent, msgs) {

	// go ahead and cache the navigator and subject info now (since it should never change)		
	this._strNav = this._getNavigatorInfo();
	this._subjPfx = this._getSubjectPrefix();

	var reportButton = new DwtDialog_ButtonDescriptor(ZmErrorDialog.REPORT_BUTTON, msgs.report, DwtDialog.ALIGN_LEFT);
	var detailButton = new DwtDialog_ButtonDescriptor(ZmErrorDialog.DETAIL_BUTTON, msgs.showDetails, DwtDialog.ALIGN_LEFT);
	DwtMessageDialog.call(this, parent, null, null, [reportButton, detailButton]);

	this.registerCallback(ZmErrorDialog.REPORT_BUTTON, this._reportCallback, this);
	this.registerCallback(ZmErrorDialog.DETAIL_BUTTON, this._showDetail, this);
	
	this._showDetailsMsg = msgs.showDetails;
	this._hideDetailsMsg = msgs.hideDetails;
}

ZmErrorDialog.prototype = new DwtMessageDialog;
ZmErrorDialog.prototype.constructor = ZmErrorDialog;

ZmErrorDialog.prototype.toString = function() {
	return "ZmErrorDialog";
};

//
// Consts
//

ZmErrorDialog.REPORT_BUTTON = ++DwtDialog.LAST_BUTTON;
ZmErrorDialog.DETAIL_BUTTON = ++DwtDialog.LAST_BUTTON;
ZmErrorDialog.REPORT_URL = "//www.zimbra.com/e/";

//
// Data
//

ZmErrorDialog.prototype._detailsVisible = false;
ZmErrorDialog.prototype.CONTROLS_TEMPLATE = "zimbra.Widgets#ZmErrorDialogControls";

//
// Public methods
//

ZmErrorDialog.prototype.reset =
function() {
	this.setDetailString();
	DwtMessageDialog.prototype.reset.call(this);
};

/**
* Sets the text that shows up when the Detail button is pressed.
*
* @param text	detail text
*/
ZmErrorDialog.prototype.setDetailString = 
function(text) {
    if (!(this._button[ZmErrorDialog.DETAIL_BUTTON])) { return; }

    this._button[ZmErrorDialog.DETAIL_BUTTON].setVisible(text != null);
    this._detailStr = text;
};

ZmErrorDialog.prototype.setMessage =
function(msgStr, detailStr, style, title) {
	this._msgStr = msgStr;
	this.setDetailString(detailStr);
	this._msgStyle = style;
	this._msgTitle = title;

	// clear the 'detailsVisible' flag and reset the title of the 'showDetails' button
	this._detailsVisible = false;
	this._button[ZmErrorDialog.DETAIL_BUTTON].setText(this._showDetailsMsg);
	
	DwtMessageDialog.prototype.setMessage.call(this, msgStr, style, title);
};


ZmErrorDialog.prototype.popdown = function() {
    DwtMessageDialog.prototype.popdown.call(this);
    this.setButtonVisible(ZmErrorDialog.REPORT_BUTTON, true);
};

//
// Protected methods
//

ZmErrorDialog.prototype._getNavigatorInfo =
function() {
	var strNav = [];
	var idx = 0;
	
	// Add the url
	strNav[idx++] = "\n\n";
	strNav[idx++] = "href: ";
	strNav[idx++] = location.href;
	strNav[idx++] = "\n";

	for (var i in navigator) {
		// Skip functions
		if(typeof navigator[i] == "function") {continue;}
		if(typeof navigator[i] == "unknown") {continue;}	// IE7
		strNav[idx++] = i + ": " + navigator[i] + "\n";
	}
	return strNav.join("");
};

ZmErrorDialog.prototype._getSubjectPrefix = 
function() {
	var strSubj = [];
	var idx = 0;
	
	strSubj[idx++] = "ER: ";
	
	if (AjxEnv.isIE) 				strSubj[idx++] = "IE ";
	else if (AjxEnv.isFirefox)		strSubj[idx++] = "FF ";
	else if (AjxEnv.isMozilla)		strSubj[idx++] = "MOZ ";
	else if (AjxEnv.isSafari)		strSubj[idx++] = "SAF ";
	else if (AjxEnv.isOpera)		strSubj[idx++] = "OPE ";
	else							strSubj[idx++] = "UKN ";

	if (AjxEnv.isWindows)			strSubj[idx++] = "WIN ";
	else if (AjxEnv.isLinux)		strSubj[idx++] = "LNX ";
	else if (AjxEnv.isMac)			strSubj[idx++] = "MAC ";
	else							strSubj[idx++] = "UNK ";

	strSubj[idx++] = appCtxt.get(ZmSetting.CLIENT_VERSION) + " ";
	return strSubj.join("");
};

ZmErrorDialog.prototype._getUserPrefs = 
function() {
	var currSearch = appCtxt.getCurrentSearch();
	var strPrefs = [];
	var idx = 0;

	// Add username and current search
	strPrefs[idx++] = "\n\n";
	strPrefs[idx++] = "username: ";
	strPrefs[idx++] = appCtxt.get(ZmSetting.USERNAME);
	strPrefs[idx++] = "\n";
	if (currSearch) {
		strPrefs[idx++] = "currentSearch: ";
		strPrefs[idx++] = currSearch.query;
		strPrefs[idx++] = "\n";
	}
	for (var i in ZmSetting.INIT) {
		if (ZmSetting.INIT[i][0]) {
			strPrefs[idx++] = ZmSetting.INIT[i][0];
			strPrefs[idx++] = ": ";
			strPrefs[idx++] = ("" + ZmSetting.INIT[i][3]);
			strPrefs[idx++] = "\n";
		}
	}
	return strPrefs.join("");
};

// Callbacks

ZmErrorDialog.prototype._reportCallback =
function() {
	// iframe initialization - recreate iframe if IE and reuse if FF
	if (!this._iframe || AjxEnv.isIE) {
		this._iframe = document.createElement("iframe");
		this._iframe.style.width = this._iframe.style.height = 0;
		this._iframe.style.visibility = "hidden";

		var contentDiv = this._getContentDiv();
		contentDiv.appendChild(this._iframe);
	}
	
	var strPrefs = this._getUserPrefs();
	var formId = Dwt.getNextId();

	// generate html form for submission via POST
	var html = [];
	var idx = 0;
	var subject = this._subjPfx + this._detailStr.substring(0,40);
	var scheme = (location.protocol == 'https:') ? "https:" : "http:";
	html[idx++] = "<html><head></head><body><form id='";
	html[idx++] = formId;
	html[idx++] = "' method='POST' action='";
	html[idx++] = scheme;
	html[idx++] = ZmErrorDialog.REPORT_URL;
	html[idx++] = "'>";
	html[idx++] = "<textarea name='details'>";
	html[idx++] = this._detailStr;
	html[idx++] = "version - ";
	html[idx++] = appCtxt.get(ZmSetting.CLIENT_VERSION);
	html[idx++] = "\n";
	html[idx++] = "release - ";
	html[idx++] = appCtxt.get(ZmSetting.CLIENT_RELEASE);
	html[idx++] = "\n";
	html[idx++] = "date - ";
	html[idx++] = appCtxt.get(ZmSetting.CLIENT_DATETIME);
	html[idx++] = "</textarea><textarea name='navigator'>";
	html[idx++] = this._strNav;
	html[idx++] = "</textarea><textarea name='prefs'>";
	html[idx++] = strPrefs;
	html[idx++] = "</textarea><textarea name='subject'>";
	html[idx++] = subject;
	html[idx++] = "</textarea></form></body></html>";

	var idoc = Dwt.getIframeDoc(this._iframe);
	idoc.open();
	idoc.write(html.join(""));
	idoc.close();

	// submit the form!
	var form = idoc.getElementById(formId);
	if (form) {
		form.submit();
	}

	this.popdown();
};

// Displays the detail text
ZmErrorDialog.prototype._showDetail = 
function() {
	this._detailsVisible = !this._detailsVisible;

	var msg = this._msgStr;
	if (this._detailsVisible) {
		msg += "<hr> " + this._detailStr.substr(0,300);
		if (this._detailStr.length > 300) msg += "...";
	}
	DwtMessageDialog.prototype.setMessage.call(this, msg, this._msgStyle, this._msgTitle);
	this._button[ZmErrorDialog.DETAIL_BUTTON].setText(this._detailsVisible ? this._hideDetailsMsg : this._showDetailsMsg);
};
}
if (AjxPackage.define("zimbra.common.ZLoginFactory")) {
/*
 * ***** BEGIN LICENSE BLOCK *****
 * 
 * Zimbra Collaboration Suite Web Client
 * Copyright (C) 2006, 2007 Zimbra, Inc.
 * 
 * The contents of this file are subject to the Yahoo! Public License
 * Version 1.0 ("License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.zimbra.com/license.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * 
 * ***** END LICENSE BLOCK *****
 */

ZLoginFactory = function() {}

ZLoginFactory.USER_ID = "ZLoginUserName";
ZLoginFactory.PASSWORD_ID = "ZLoginPassword";
ZLoginFactory.REMEMBER_ME_ID = "rememberMe";
ZLoginFactory.REMEMBER_ME_CONTAINER_ID = "ZLoginRememberMeContainer"
ZLoginFactory.NEW_PASSWORD_ID = "newpass1";
ZLoginFactory.NEW_PASSWORD_TR_ID = "ZLoginNewPassword1Tr";
ZLoginFactory.PASSWORD_CONFIRM_TR_ID = "ZLoginNewPassword2Tr";
ZLoginFactory.PASSWORD_CONFIRM_ID = "newpass2";
ZLoginFactory.LOGIN_BUTTON_ID = "ZLoginButton";
ZLoginFactory.HIDDEN_BUTTON_ID = "ZLoginHiddenButton";

// Constants for tabbing through the login controls.
ZLoginFactory.TEXT_TYPE = 0;
ZLoginFactory.CHECKBOX_TYPE = 1;
ZLoginFactory.BUTTON_TYPE = 2;

ZLoginFactory.TAB_ORDER = [ZLoginFactory.USER_ID, ZLoginFactory.PASSWORD_ID, 
					 ZLoginFactory.NEW_PASSWORD_ID, ZLoginFactory.PASSWORD_CONFIRM_ID,
					 ZLoginFactory.REMEMBER_ME_ID, ZLoginFactory.LOGIN_BUTTON_ID];
ZLoginFactory.VISIBILITY = [ZLoginFactory.USER_ID, ZLoginFactory.PASSWORD_ID, 
					  ZLoginFactory.NEW_PASSWORD_TR_ID, ZLoginFactory.PASSWORD_CONFIRM_TR_ID,
					  ZLoginFactory.REMEMBER_ME_CONTAINER_ID, ZLoginFactory.LOGIN_BUTTON_ID];
ZLoginFactory.TAB_TYPE = [ZLoginFactory.TEXT_TYPE, ZLoginFactory.TEXT_TYPE, 
					ZLoginFactory.TEXT_TYPE, ZLoginFactory.TEXT_TYPE,
					ZLoginFactory.CHECKBOX_TYPE, ZLoginFactory.BUTTON_TYPE];
				
/*
 * Creates a copy of the default login parameters.
 * 
 * @param msgs	The class where localized messages are defined. ZmMsg for example.
 */
ZLoginFactory.copyDefaultParams = 
function(msgs) {
	return {
		showPanelBorder: true,
		
		companyURL : msgs["splashScreenCompanyURL"] || "",
	
		shortVersion : "",
		longVersion : "",
	
		appName : msgs["splashScreenAppName"] || "",
		productName : "",
	
		showError : false,
		errorMsg : "",
	
		showAbout : false,
		aboutMsg : "",
	
		showLoading : false,
		loadingMsg : msgs["splashScreenLoading"] || "",
		
		showForm : false,
		
		showUserField : false,
		userNameMsg : msgs["username"] ? msgs["username"] + ':' : "",
		
		showPasswordField : false,
		passwordMsg : msgs["password"] ? msgs["password"] + ':' : "",
		
		showNewPasswordFields : false,
		newPassword1Msg : msgs["newPassword"] + ':'|| "",
		newPassword2Msg : msgs["confirm"] + ':'|| "",
	
		showLicenseMsg : false,
		licenseMsg : "",
		
		showRememberMeCheckbox : false,
		rememberMeMsg : msgs["rememberMe"] || "",
	
		showLogOff : false,
		logOffMsg : msgs["loginAsDiff"] || "",
		logOffAction : "",
		
		showButton : false,
		buttonName : msgs["login"] || "",
		
		copyrightText : msgs["splashScreenCopyright"] || ""
	};
};

// show and hide various things
ZLoginFactory.getLoginPanel = function () 			{												
	var retval = this.get("ZLoginPanel");
	if(!retval)
		retval = this.get("ZLoginPanelInvisible");

	return retval;
	
}

ZLoginFactory.showErrorMsg = function (msg) {
	this.setHTML("ZLoginErrorMsg", msg);
	this.show("ZLoginErrorPanel");
	this._flickerErrorMessagePanel();
}
ZLoginFactory.hideErrorMsg = function () 			{												this.hide("ZLoginErrorPanel");	}
ZLoginFactory.getErrorMsgPanel = function () 		{												return this.get("ZLoginErrorPanel");	}

ZLoginFactory.showAboutMsg = function (msg) 		{	this.setHTML("ZLoginAboutPanel", msg);		this.show("ZLoginAboutPanel");	}
ZLoginFactory.hideAboutMsg = function () 			{												this.hide("ZLoginAboutPanel");	}
ZLoginFactory.getAboutMsg = function () 			{												return this.get("ZLoginAboutPanel");	}

ZLoginFactory.showLoadingMsg = function (msg)		{	this.setHTML("ZLoginLoadingMsg", msg);		this.show("ZLoginAboutPanel");	}
ZLoginFactory.hideLoadingMsg = function () 		{													this.hide("ZLoginAboutPanel");	}
ZLoginFactory.getLoadingMsg = function () 		{													return this.get("ZLoginAboutPanel");	}

ZLoginFactory.showForm = function ()				{												this.show("ZLoginFormPanel");	}
ZLoginFactory.hideForm = function () 				{												this.hide("ZLoginFormPanel");	}
ZLoginFactory.getForm = function () 				{												return this.get("ZLoginFormPanel");	}

ZLoginFactory.showUserField = function (name)		{	this.setValue(ZLoginFactory.USER_ID, name);				this.show(ZLoginFactory.USER_ID);	}
ZLoginFactory.hideUserField = function () 			{												this.hide(ZLoginFactory.USER_ID);	}
ZLoginFactory.getUserField = function () 			{												return this.get(ZLoginFactory.USER_ID);	}

ZLoginFactory.showPasswordField = function (msg)	{	this.show(ZLoginFactory.PASSWORD_ID);	}
ZLoginFactory.hidePasswordField = function () 		{	this.hide(ZLoginFactory.PASSWORD_ID);	}
ZLoginFactory.getPasswordField = function () 		{	return this.get(ZLoginFactory.PASSWORD_ID);	}

ZLoginFactory.showNewPasswordFields = function ()	{	this.show(ZLoginFactory.NEW_PASSWORD_TR_ID); this.show(ZLoginFactory.PASSWORD_CONFIRM_TR_ID);	}
ZLoginFactory.hideNewPasswordFields = function () 	{	this.hide(ZLoginFactory.NEW_PASSWORD_TR_ID); this.hide(ZLoginFactory.PASSWORD_CONFIRM_TR_ID);	}
ZLoginFactory.areNewPasswordFieldsShown = function (){	return this.isShown(ZLoginFactory.NEW_PASSWORD_TR_ID); }

ZLoginFactory.getNewPasswordField = function () 	{	return this.get(ZLoginFactory.NEW_PASSWORD_ID); }
ZLoginFactory.getPasswordConfirmField = function () {	return this.get(ZLoginFactory.PASSWORD_CONFIRM_ID); }

ZLoginFactory.showRememberMeCheckbox = function ()	{	this.show(ZLoginFactory.REMEMBER_ME_CONTAINER_ID);	}
ZLoginFactory.hideRememberMeCheckbox = function ()	{	this.hide(ZLoginFactory.REMEMBER_ME_CONTAINER_ID);	}

ZLoginFactory.showLogOff = function ()	{	this.show("ZLoginLogOffContainer");	}
ZLoginFactory.hideLogOff = function ()	{	this.hide("ZLoginLogOffContainer");	}

ZLoginFactory.setLoginButtonName = function (name) 	{	this.setHTML("ZLoginButtonText", name);	}
ZLoginFactory.setLoginButtonAction = function (method) {	var el = document.getElementById(ZLoginFactory.LOGIN_BUTTON_ID); if (el) el.onclick = method	}
ZLoginFactory.getLoginButton = function () 		{	return this.get(ZLoginFactory.LOGIN_BUTTON_ID);	}


ZLoginFactory.getLoginDialogHTML = function (params) {
	var html = [
		 "<div ", (params.showPanelBorder ? "id='ZLoginPanel'" :  "id='ZLoginPanelInvisible'"), ">",
			"<table class='zLoginTable' width='100%' cellpadding=0 cellspacing=0>",
				"<tr><td id='ZLoginHeaderContainer'><center>",
						"<table class='zLoginTable'>",
							"<tr><td id='ZLoginBannerContainer'>",
									"<div id='ZLoginBannerPanel'>",
										"<table class='zLoginTable'><tr>",
											"<td><div style='cursor:pointer' id='ZLoginBannerImage' class='ImgLoginBanner'", 
													" onclick='window.open(\"", params.companyURL, "\", \"_blank\")'></div></td>",
											"<td valign=top id='ZLoginShortVersion'>", params.shortVersion, "</td>",
										"</tr></table>",
										"<div id='ZLoginAppName'>", params.appName, "</div>",
										"<div id='ZLoginProductName'>", params.productName, "</div>",
										"<div id='ZLoginLongVersion'>", params.longVersion, "</div>",
									"</div>",
								"</td>",
							"</tr>",
						"</table>",
					"</center></td>",
				"</tr>",
				"<tr><td id='ZLoginBodyContainer'>",
						"<div id='ZLoginErrorPanel' ", (params.showError ? "" : "style='display:none'"), ">",
							"<table>",
								"<tr><td valign='top' width='40'><div class='ImgCritical_32'></td>",
									"<td width='*' id='ZLoginErrorMsg' class='errorText'>", params.errorMsg, "</td>",
								"</tr>",
							"</table>",
						"</div>",
						"",
						"<div id='ZLoginAboutPanel' ", (params.showAbout ? "" : "style='display:none'"), ">", params.aboutMsg,
						"</div>",
						"<div id='ZLoginLoadingPanel' ", (params.showLoading ? "" : "style='display:none'"), ">",
							"<table><tr><td></td><td id='ZLoginLoadingMsg'>", params.loadingMsg, "</td></tr></table>",
						"</div>",
						"<div id='ZLoginFormPanel' ", (params.showForm ? "" : "style='display:none'"), ">",
							"<table class='zLoginTable' width='100%' cellpadding=4>",
								"<tr id='ZLoginUserTr'", (params.showUserField ? "" : "style='display:none'"), "><td class='zLoginLabelContainer'>", params.userNameMsg, "</td>",
									"<td class='zLoginFieldContainer' colspan=2>",
										Dwt.CARET_HACK_BEGIN, "<input id='", ZLoginFactory.USER_ID, "' class='zLoginField' autocomplete=OFF type=text tabIndex=1>", Dwt.CARET_HACK_END,
									"</td>",
								"</tr>",
								"<tr id='ZLoginPasswordTr' ", (params.showPasswordField ? "" : "style='display:none'"), "><td class='zLoginLabelContainer'>", params.passwordMsg, "</td>",
									"<td class='zLoginFieldContainer' colspan=2>",
										Dwt.CARET_HACK_BEGIN, "<input id='", ZLoginFactory.PASSWORD_ID, "' class='zLoginField' autocomplete=OFF type=password tabIndex=2>", Dwt.CARET_HACK_END, 
									"</td>",
								"</tr>",
								"<tr id='", ZLoginFactory.NEW_PASSWORD_TR_ID, "' ", (params.showNewPasswordFields ? "" : "style='display:none'"), "><td class='zLoginLabelContainer'>", params.newPassword1Msg, "</td>",
									"<td class='zLoginFieldContainer' colspan=2>",
										Dwt.CARET_HACK_BEGIN, "<input id='", ZLoginFactory.NEW_PASSWORD_ID, "' class='zLoginField' autocomplete=OFF type=password tabIndex=2>", Dwt.CARET_HACK_END, 
									"</td>",
								"</tr>",
								"<tr id='", ZLoginFactory.PASSWORD_CONFIRM_TR_ID, "' ", (params.showNewPasswordFields ? "" : "style='display:none'"), "><td class='zLoginLabelContainer'>", params.newPassword2Msg, "</td>",
									"<td class='zLoginFieldContainer' colspan=2>",
										Dwt.CARET_HACK_BEGIN, "<input id='", ZLoginFactory.PASSWORD_CONFIRM_ID, "' class='zLoginField' autocomplete=OFF type=password tabIndex=2>", Dwt.CARET_HACK_END, 
									"</td>",
								"</tr>",
								"<tr id='ZLoginLicenseMsgContainer' ", (params.showLicenseMsg ? "" : "style='display:none'"), ">",
									"<td colspan=3 id='ZLoginLicenseMsg'>", params.licenseMsg, "</td>",
								"</tr>",
								"<tr><td class='zLoginLabelContainer'>&nbsp;</td>",
									"<td class='zLoginFieldContainer' id='", ZLoginFactory.REMEMBER_ME_CONTAINER_ID, "' ", (params.showRememberMeCheckbox ? "" : "style='display:none'"), ">",
										"<table class='zLoginTable' width=100%>",
											"<tr><td width=1><input id='", ZLoginFactory.REMEMBER_ME_ID, "' type='checkbox'></td>",
												"<td class='zLoginCheckboxLabelContainer'>", params.rememberMeMsg, "</td>",
											"</tr>",
										"</table>",
									"</td>",
									"<td class='zLoginFieldContainer' id='ZLoginLogOffContainer' ", (params.showLogOff ? "" : "style='display:none'"), ">",
										"<a href='' onclick='javascript:" + params.logOffAction + ";'>" + params.logOffMsg + "</a>",
									"</td>",
									"<td class='zLoginButtonContainer' align='right'", (params.showButton ? "" : "style='display:none'"), ">",
										"<div id='", ZLoginFactory.LOGIN_BUTTON_ID, "' class='DwtButton'",
											"onclick='", params.loginAction, ";return false'",
											"onmouseover='javascript:this.className=\"DwtButton-hover\"'",
											"onmouseout='javascript:this.className=\"DwtButton\"'",
											"onmousedown='javascript:this.className=\"DwtButton-active\";return false'",
											"onmouseup='javascript:this.className=\"DwtButton\"'",
											"onmousemove='javascript:return false'",
											"onselectstart='javascript: return false'",
											"onfocus='javascript:this.className = \"DwtButton-focused\";return false'",
											"onblur='javascript:this.className = \"DwtButton\";return false'",
										">",
											"<table style='width:100%;height:100%' cellspacing=0>",
												"<tr><td align='center' class='Text' id='ZLoginButtonText'>",
														params.buttonName,
												"</td></tr>",
											"</table>",
										"</div>",
										"<!-- non-IE browsers dont allow focus for non-INPUT elements so we have to",
											" create a hidden input to fake focus for our DIV which acts as an input button -->",
										"<input type='button' style='position:absolute;top:-10000;left:-10000;' id='", ZLoginFactory.HIDDEN_BUTTON_ID, "'>",
									"</td>",
								"</tr>",
							"</table>",
						"</div>",
						"<table class='zLoginTable' width='100%' cellpadding=0 cellspacing=0>",
							"<tr>",
								"<td id='ZloginClientLevelContainer'>",
									params.clientLevelNotice,
								"</td>",
							"</tr>",
							"<tr>",
								"<td id='ZLoginLicenseContainer'>",
									params.copyrightText,
								"</td>",
							"</tr>",
						"</table>",
					"</td>",
				"</tr>",
			"</table>",
		 "</div>"
	].join("");
	return html;
}


// simple API to show/hide elements (can be replaced with Dwt if desired)
ZLoginFactory.setHTML = function (id, newContent) {
	var el = document.getElementById(id);
	if (el && newContent != null) el.innerHTML = newContent;
}

ZLoginFactory.setValue = function (id, newContent) {
	var el = document.getElementById(id);
	if (el && newContent != null) el.value = newContent;
}

ZLoginFactory.show = function (id, newContent) {
	var el = document.getElementById(id);
	if (el) el.style.display = "";
}

ZLoginFactory.isShown = function (id) {
	var el = document.getElementById(id);
	return el ? (el.style.display != "none") : false;
}

ZLoginFactory.hide = function (id) {
	var el = document.getElementById(id);
	if (el) el.style.display = "none";
}

ZLoginFactory.get = function (id) {
	return document.getElementById(id);
}

ZLoginFactory.handleKeyPress =
function(ev) {
    ev = ev || window.event;
    if (ev == null) {
    	return true;
    }
    var target = ev.target ? ev.target: ev.srcElement;
    if (!target) {
    	return true;
    }
    var keyCode = ev.keyCode;
    var fakeTabKey = false;
    if (keyCode == 13) { // Enter
		if (target.id == ZLoginFactory.USER_ID || target.id == ZLoginFactory.NEW_PASSWORD_ID) {
			fakeTabKey = true;
		} else {
			// Call the login action
			var loginAction = ZLoginFactory.get(ZLoginFactory.LOGIN_BUTTON_ID).onclick;
			if (loginAction) {
				loginAction.call(target);
			}
			ZLoginFactory._cancelEvent(ev);
			return false;
		}
    }
	if (fakeTabKey || (keyCode == 9)) { // Tab
		var startIndex = ZLoginFactory.TAB_ORDER.length - 1;
		for (var i = 0; i < ZLoginFactory.TAB_ORDER.length; i++) {
			if (ZLoginFactory.TAB_ORDER[i] == target.id) {
				startIndex = i;
				break;
			}
		}
		var forward = !ev.shiftKey;
		var tabToIndex = ZLoginFactory._getTabToIndex(startIndex, forward);
		var tabToId = ZLoginFactory.TAB_ORDER[tabToIndex];
		var tabToType = ZLoginFactory.TAB_TYPE[tabToIndex];
		ZLoginFactory._onFocusChange(tabToType, tabToId, target);
		ZLoginFactory._cancelEvent(ev);
	}
}

// Private / protected methods

ZLoginFactory._cancelEvent =
function(ev) {
	if (ev.stopPropagation)
		ev.stopPropagation();

	if (ev.preventDefault)
		ev.preventDefault();

	ev.cancelBubble = true;
	ev.returnValue = false;
}

ZLoginFactory._onFocusChange =
function(type, id, target) {
	if ((type != ZLoginFactory.BUTTON_TYPE) && !AjxEnv.isIE) {
		ZLoginFactory._loginButtonBlur();
	}	
	if (type == ZLoginFactory.TEXT_TYPE) {
		var edit = ZLoginFactory.get(id);
		edit.focus();
		edit.select();
	} else if (type == ZLoginFactory.CHECKBOX_TYPE) {
		var checkbox = ZLoginFactory.get(id);
		checkbox.focus();
	}
	else {
		var button = ZLoginFactory.get(id);
		if (AjxEnv.isIE) {
			button.focus();
		} else {
			ZLoginFactory._loginButtonFocus(button);
			target.blur();
			document.getElementById(ZLoginFactory.HIDDEN_BUTTON_ID).focus();
		}
	}
};

ZLoginFactory._getTabToIndex =
function(startIndex, forward) {
	var testIndex = startIndex;
	do {
		var tabToIndex;
		if (forward) {
			testIndex = (testIndex == (ZLoginFactory.TAB_ORDER.length - 1)) ? 0 : testIndex + 1;
		} else {
			testIndex = (testIndex == 0) ? (ZLoginFactory.TAB_ORDER.length - 1) : testIndex - 1;
		}
		var id = ZLoginFactory.TAB_ORDER[testIndex];
		var visibilityId = ZLoginFactory.VISIBILITY[testIndex];
		var control = ZLoginFactory.get(id);
		if (ZLoginFactory.isShown(visibilityId) && !ZLoginFactory.get(id).disabled) {
			return testIndex
		}
	} while (testIndex != startIndex);
	return 0; // Should never get here.
}
					 
ZLoginFactory._loginButtonFocus =
function(border) {
	border.className = "DwtButton-focused";
};

ZLoginFactory._loginButtonBlur =
function(button) {
	var button = ZLoginFactory.get(ZLoginFactory.LOGIN_BUTTON_ID);
	button.className = "DwtButton";
};

/*
* Hide error panel very briefly, making it look like something happened if
* user has successive errors.
*/
ZLoginFactory._flickerErrorMessagePanel =
function() {
	ZLoginFactory.getErrorMsgPanel().style.visibility = "hidden";
	window.setTimeout(ZLoginFactory._showErrorMessagePanel, 8);
};

ZLoginFactory._showErrorMessagePanel =
function() {
	ZLoginFactory.getErrorMsgPanel().style.visibility = "visible";
};
}
}
if (AjxPackage.define("zimbra.Widgets")) {
AjxTemplate.register("zimbra.Widgets#ZmErrorDialogControls", 
function(name, params, data, buffer) {
	var _hasBuffer = Boolean(buffer);
	data = (typeof data == "string" ? { id: data } : data) || {};
	buffer = buffer || [];
	var _i = buffer.length;

	buffer[_i++] = "<div id='";
	buffer[_i++] = data["id"];
	buffer[_i++] = "_buttons' class='DwtDialogButtonBar'>";
	 if (AjxEnv.isNav) { 
	buffer[_i++] = "<input type='button' id='";
	buffer[_i++] = data["id"];
	buffer[_i++] = "_focus' style='height:0px;width:0px;display:none;'>";
	 } 
	buffer[_i++] = "</div>";

	return _hasBuffer ? buffer.length : buffer.join("");
},
{
	"width": "20",
	"height": "32",
	"id": "zimbra.Widgets#ZmErrorDialogControls"
}, false);
AjxTemplate.register("zimbra.Widgets", AjxTemplate.getTemplate("zimbra.Widgets#ZmErrorDialogControls"), AjxTemplate.getParams("zimbra.Widgets#ZmErrorDialogControls"));

}
